from glob import glob
import os
from learning_loop_node import loop
from urllib.parse import urljoin
from requests import Session
import asyncio


class LiveServerSession(Session):
    """https://stackoverflow.com/a/51026159/364388"""

    def __init__(self, *args, **kwargs):
        super(LiveServerSession, self).__init__(*args, **kwargs)
        self.prefix_url = 'https://preview.learning-loop.ai'

    def request(self, method, url, *args, **kwargs):
        url = urljoin(self.prefix_url, url)
        headers = {}
        if 'token' in url:
            return super(LiveServerSession, self).request(method, url, *args, **kwargs)

        eventloop = asyncio.get_event_loop()
        headers = eventloop.run_until_complete(loop.instance.get_headers())
        return super(LiveServerSession, self).request(method, url, headers=headers, *args, **kwargs)


def get_files_in_folder(folder: str):
    files = [entry for entry in glob(f'{folder}/**/*', recursive=True) if os.path.isfile(entry)]
    files.sort()
    return files



import asyncio
import traceback
from fastapi_utils.tasks import repeat_every
from fastapi.encoders import jsonable_encoder
from typing import Union
from uuid import uuid4
from icecream import ic
from .model import Model
from .trainer import Trainer
from learning_loop_node.status import TrainingStatus
from learning_loop_node.node import Node, State
import logging

class TrainerNode(Node):
    trainer: Trainer
    latest_known_model_id: Union[str, None]
    skip_check_state: bool = False

    def __init__(self, name: str, uuid: str, trainer: Trainer):
        super().__init__(name, uuid)
        self.trainer = trainer
        self.latest_known_model_id = None

        @self.sio_client.on('begin_training')
        async def on_begin_training(organization, project, source_model):
            loop = asyncio.get_event_loop()
            loop.create_task(self.begin_training(organization, project, source_model))
            return True

        @self.sio_client.on('stop_training')
        async def stop():
            return await self.stop_training()

        @self.sio_client.on('save')
        def on_save(organization, project, model):
            loop = asyncio.get_event_loop()
            loop.create_task(self.save_model(organization, project, model['id']))
            return True

        @self.on_event("startup")
        @repeat_every(seconds=5, raise_exceptions=True, wait_first=False)
        async def check_state():
            if not self.skip_check_state:
                await self.check_state()

    async def begin_training(self, organization: str, project: str, source_model: dict):
        self.status.latest_error = None
        await self.update_state(State.Preparing)
        try:
            await self.trainer.begin_training(organization, project, source_model)
        except Exception as e:
            traceback.print_exc()
            self.status.latest_error = f'Could not start training: {str(e)})'
            self.trainer.stop_training()
            await self.update_state(State.Idle)
            return
        self.latest_known_model_id = source_model['id']
        await self.update_state(State.Running)

    async def stop_training(self) -> Union[bool, str]:
        try:
            self.trainer.stop_training()
            self.trainer.training = None
            await self.update_state(State.Idle)
        except Exception as e:
            self.status.latest_error = f'Could not stop training: {str(e)})'
            traceback.print_exc()
            await self.send_status()

            return False
        self.latest_known_model_id = None
        await self.send_status()
        return True

    async def save_model(self, organization, project, model_id):
        try:
            await self.trainer.save_model(organization, project, model_id)
        except Exception as e:
            traceback.print_exc()
            await self.update_error_msg(f'Could not save model: {str(e)}')

    async def check_state(self):
        logging.debug(f'training {"none" if self.trainer.training == None else "running"}, state: {self.status.state}')
        try:
            if self.status.state == State.Running and not self.trainer.is_training_alive():
                raise Exception()
        except:
            await self.update_error_msg(f'Training crashed.')

        await self.try_get_new_model()

    async def try_get_new_model(self) -> None:
        try:
            current_training = self.trainer.training
            if self.status.state == State.Running and current_training:
                model = self.trainer.get_new_model()
                if model:
                    new_model = Model(
                        id=str(uuid4()),
                        confusion_matrix=model.confusion_matrix,
                        parent_id=self.latest_known_model_id,
                        train_image_count=self.trainer.training.data.train_image_count(),
                        test_image_count=self.trainer.training.data.test_image_count(),
                        trainer_id=self.uuid,
                    )

                    result = await self.sio_client.call('update_model', (current_training.context.organization, current_training.context.project, jsonable_encoder(new_model)))
                    if result != True:
                        await self.update_error_msg(f'Could not update_model: {result}')
                        return

                    logging.info(f'successfully uploaded model {jsonable_encoder(new_model)}')
                    self.trainer.on_model_published(model, new_model.id)
                    self.latest_known_model_id = new_model.id
                    await self.send_status()
        except Exception as e:
            await self.update_error_msg(f'Could not get new model: {str(e)}')

    async def send_status(self):
        status = TrainingStatus(
            id=self.uuid,
            name=self.name,
            state=self.status.state,
            uptime=self.status.uptime,
            latest_produced_model_id=self.latest_known_model_id,
            latest_error=self.status.latest_error
        )

        print('sending status', status, flush=True)
        result = await self.sio_client.call('update_trainer', jsonable_encoder(status), timeout=1)
        if not result == True:
            raise Exception(result)
        print('status send', flush=True)

    async def update_error_msg(self, msg: str) -> None:
        self.status.latest_error = msg
        await self.send_status()



from learning_loop_node.trainer.capability import Capability
from learning_loop_node.trainer.downloader_factory import DownloaderFactory
import pytest
from learning_loop_node.tests import test_helper
from learning_loop_node.trainer.tests import trainer_test_helper
from learning_loop_node.trainer.downloader import DataDownloader
from learning_loop_node import node, node_helper
from learning_loop_node.context import Context
from learning_loop_node.conftest import create_project
from icecream import ic
import asyncio


@pytest.fixture(autouse=True, scope='module')
def create_project_for_module():
    # TODO can we use the 'create_project' fixture here?
    test_helper.LiveServerSession().delete(f"/api/zauberzeug/projects/pytest?keep_images=true")
    project_configuration = {'project_name': 'pytest', 'inbox': 0, 'annotate': 0, 'review': 0, 'complete': 3, 'image_style': 'beautiful',
                             'categories': 2, 'thumbs': False, 'tags': 0, 'trainings': 1, 'detections': 3, 'annotations': 0, 'skeleton': False}
    assert test_helper.LiveServerSession().post(f"/api/zauberzeug/projects/generator",
                                                json=project_configuration).status_code == 200
    yield
    test_helper.LiveServerSession().delete(f"/api/zauberzeug/projects/pytest?keep_images=true")


@pytest.fixture
def downloader() -> DataDownloader:
    context = Context(organization='zauberzeug', project='pytest')
    return DownloaderFactory.create(context=context, capability=Capability.Box)


@pytest.mark.asyncio
async def test_download_model():
    _, _, trainings_folder = trainer_test_helper.create_needed_folders()
    model_id = await trainer_test_helper.assert_upload_model()

    await node_helper.download_model(trainings_folder,
                                     'zauberzeug', 'pytest', model_id, 'mocked')

    files = test_helper.get_files_in_folder('/data')
    assert len(files) == 2

    assert files[0] == "/data/zauberzeug/pytest/trainings/some_uuid/file_1.txt"
    assert files[1] == "/data/zauberzeug/pytest/trainings/some_uuid/file_2.txt"

    assert open(files[0], 'r').read() == 'content of file one'
    assert open(files[1], 'r').read() == 'content of file two'


@pytest.mark.asyncio
async def test_download_basic_data(downloader: DataDownloader):
    basic_data = await downloader.download_basic_data()

    assert len(basic_data.image_ids) == 3
    assert len(basic_data.box_categories) == 2


@pytest.mark.asyncio
async def test_download_images(downloader: DataDownloader):
    _, image_folder, _ = trainer_test_helper.create_needed_folders()

    basic_data = await downloader.download_basic_data()
    await downloader.download_images(asyncio.get_event_loop(), basic_data.image_ids, image_folder)
    files = test_helper.get_files_in_folder('/data')

    assert len(files) == 3


@pytest.mark.asyncio
async def test_download_training_data(downloader: DataDownloader):
    _, image_folder, _ = trainer_test_helper.create_needed_folders()
    basic_data = await downloader.download_basic_data()
    image_data = await downloader.download_image_data(basic_data.image_ids)

    assert len(image_data) == 3



from learning_loop_node.trainer.downloader_factory import DownloaderFactory
from learning_loop_node import node_helper
from typing import List, Optional
from uuid import uuid4
import os
from pydantic.main import BaseModel
from learning_loop_node.trainer.capability import Capability
from learning_loop_node.trainer.training import Training
from learning_loop_node.trainer.model import BasicModel
from learning_loop_node.context import Context
from learning_loop_node.node import Node
from icecream import ic


class Trainer(BaseModel):
    training: Optional[Training]
    capability: Capability
    model_format: str

    async def begin_training(self, organization: str, project: str, source_model: dict) -> None:
        context = Context(organization=organization, project=project)
        downloader = DownloaderFactory.create(context, self.capability)

        self.training = Trainer.generate_training(context, source_model)
        self.training.data = await downloader.download_data(self.training.images_folder)

        await node_helper.download_model(self.training.training_folder,
                                         organization, project, source_model['id'], self.model_format)

        await self.start_training()

    async def start_training(self) -> None:
        raise NotImplementedError()

    def stop_training(self) -> None:
        raise NotImplementedError()

    def is_training_alive(self) -> bool:
        raise NotImplementedError()

    async def save_model(self,  organization: str, project: str, model_id) -> None:
        files = self.get_model_files(model_id)
        await node_helper.upload_model(organization, project, files, model_id, self.model_format)

    def get_model_files(self, model_id) -> List[str]:
        raise NotImplementedError()

    def get_new_model(self) -> Optional[BasicModel]:
        raise NotImplementedError()

    def on_model_published(self, basic_model: BasicModel, uuid: str) -> None:
        raise NotImplementedError()

    @staticmethod
    def generate_training(context: Context, source_model: dict) -> Training:
        training_uuid = str(uuid4())
        project_folder = Node.create_project_folder(context.organization, context.project)
        return Training(
            id=training_uuid,
            base_model=source_model,
            context=context,
            project_folder=project_folder,
            images_folder=Trainer.create_image_folder(project_folder),
            training_folder=Trainer.create_training_folder(project_folder, training_uuid)
        )

    @staticmethod
    def create_image_folder(project_folder: str) -> str:
        image_folder = f'{project_folder}/images'
        os.makedirs(image_folder, exist_ok=True)
        return image_folder

    @staticmethod
    def create_training_folder(project_folder: str, trainings_id: str) -> str:
        training_folder = f'{project_folder}/trainings/{trainings_id}'
        os.makedirs(training_folder, exist_ok=True)
        return training_folder



import asyncio
from datetime import datetime, timedelta
from async_generator import asynccontextmanager
from aiohttp.client_reqrep import ClientResponse
import aiohttp
import os
import werkzeug
from icecream import ic
import logging


class AccessToken():
    def __init__(self, token: str, local_expire_date: datetime):
        self.token = token
        self.local_expire_date = local_expire_date

    def is_still_valid(self) -> bool:
        return (self.local_expire_date - timedelta(hours=1)) > datetime.now()

    def is_invalid(self) -> bool:
        return not self.is_still_valid

async def token_from_response(response: ClientResponse) -> AccessToken:
    content = await response.json()
    server_time = werkzeug.http.parse_date(response.headers['date'])
    expires_time = datetime.fromtimestamp(content['expires'], tz=server_time.tzinfo)
    # we do not care about the few seconds wich ellapsed during request.
    local_expire_time = datetime.now() + (expires_time - server_time)
    return AccessToken(content['access_token'], local_expire_time)


class Loop():
    def __init__(self) -> None:
        self.base_url: str = 'http://' + os.environ.get('HOST', 'learning-loop.ai')
        self.username: str = os.environ.get('USERNAME', None)
        self.password: str = os.environ.get('PASSWORD', None)
        self.access_token = None

    async def download_token(self):
      
        credentials = {
            'username': self.username,
            'password': self.password,
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(f'{self.base_url}/api/token', data=credentials) as response:
                logging.debug(f'received token from {self.base_url}')
                assert response.status == 200

                return await token_from_response(response)
    
    async def get_headers(self) -> dict:
        headers = {}
        if self.username is None:
            return headers

        if self.access_token is None or self.access_token.is_invalid():
            self.access_token = await self.download_token()

        headers['Authorization'] = f'Bearer {self.access_token.token}'
        return headers
        

    @asynccontextmanager
    async def get(self, path):
        headers = await self.get_headers()
        async with aiohttp.ClientSession() as session:
            async with session.get(f'{self.base_url}/{path}', headers=headers) as response:
                yield response

    @asynccontextmanager
    async def put(self, path, data):
        headers = await self.get_headers()
        async with aiohttp.ClientSession() as session:
            async with session.put(f'{self.base_url}/{path}', headers=headers, data=data) as response:
                yield response

    @asynccontextmanager
    async def post(self, path, data):
        headers = await self.get_headers()
        async with aiohttp.ClientSession() as session:
            async with session.post(f'{self.base_url}/{path}', headers=headers, data=data) as response:
                yield response


loop = Loop()



from typing import List, Optional
from learning_loop_node.node import Node
from learning_loop_node.trainer.trainer import Trainer
import aiohttp
from learning_loop_node import loop


async def assert_upload_model(file_paths: Optional[List[str]] = None) -> str:
    if not file_paths:
        file_paths = ['learning_loop_node/trainer/tests/test_data/file_1.txt',
                      'learning_loop_node/trainer/tests/test_data/file_2.txt']
    data = [('files', open(path, 'rb')) for path in file_paths]

    data = aiohttp.FormData()

    for path in file_paths:
        data.add_field('files',  open(path, 'rb'))
    async with loop.LoopHttp().post('api/zauberzeug/projects/pytest/models/mocked', data) as response:
        assert response.status == 200
        return (await response.json())['id']


def create_needed_folders(training_uuid='some_uuid'):
    project_folder = Node.create_project_folder('zauberzeug', 'pytest')
    image_folder = Trainer.create_image_folder(project_folder)
    training_folder = Trainer.create_training_folder(project_folder, training_uuid)
    return project_folder, image_folder, training_folder



import logging
import aiohttp
from .status import Status, State
from fastapi import FastAPI
import socketio
import asyncio
import asyncio
import os
from icecream import ic
from .loop import loop
from aiohttp.client_exceptions import ClientConnectorError
import logging


class Node(FastAPI):
    name: str
    uuid: str

    def __init__(self, name: str, uuid: str):
        super().__init__()
        self.ws_url = 'ws://' + os.environ.get('HOST', 'learning-loop.ai')
        self.organization = os.environ.get('ORGANIZATION', None)
        self.project = os.environ.get('PROJECT', None)

        self.name = name
        self.uuid = uuid

        self.sio_client = socketio.AsyncClient(
            reconnection_delay=0,
            request_timeout=0.5,
            # logger=True, engineio_logger=True
        )
        self.reset()

        @self.sio_client.on('connect')
        async def on_connect():
            logging.debug('received "on_connect" from constructor event.')
            self.reset()
            await self.update_state(State.Idle)

        @self.sio_client.on('disconnect')
        async def on_disconnect():
            logging.debug('received "on_disconnect" from constructor event.')
            await self.update_state(State.Offline)

        self.register_lifecycle_events()

    def register_lifecycle_events(self):
        @self.on_event("startup")
        async def startup():
            logging.debug('received "startup" event')
            await self.connect()

        @self.on_event("shutdown")
        async def shutdown():
            logging.debug('received "shutdown" event')
            await self.sio_client.disconnect()

    def reset(self):
        self.status = Status(id=self.uuid, name=self.name)

    async def connect(self):
        try:
            await self.sio_client.disconnect()
        except:
            pass

        logging.info(f'connecting to Learning Loop at {self.ws_url}')
        try:
            headers = await loop.get_headers()
            await self.sio_client.connect(f"{self.ws_url}", headers=headers, socketio_path="/ws/socket.io")
            logging.debug(f'my sid is {self.sio_client.sid}')
            logging.info('connected to Learning Loop')
        except socketio.exceptions.ConnectionError as e:
            logging.exception(f'socket.io connection error to "{self.ws_url}"')
            if not ('Already connected' in str(e) or 'Connection refused' in str(e) or 'Unexpected status code' in str(e)):
                await asyncio.sleep(0.5)
                await self.connect()
        except ConnectionRefusedError or ClientConnectorError:
            await asyncio.sleep(0.5)
            await self.connect()
        except Exception:
            logging.exception(f'error while connecting to "{self.ws_url}"')
            await asyncio.sleep(2)
            await self.connect()

    async def update_state(self, state: State):
        self.status.state = state
        if self.status.state != State.Offline:
            await self.send_status()

    async def update_status(self, new_status: Status):
        self.status.id = new_status.id
        self.status.name = new_status.name
        self.status.uptime = new_status.uptime
        self.status.latest_error = new_status.latest_error

        if self.status.state != State.Offline:
            self.status.state = State.Idle
        await self.send_status()

    async def send_status(self):
        raise Exception("Override this in subclass")

    @staticmethod
    def create_project_folder(organization: str, project: str) -> str:
        project_folder = f'/data/{organization}/{project}'
        os.makedirs(project_folder, exist_ok=True)
        return project_folder



import aiohttp
from typing import List, Tuple
import shutil
from io import BytesIO
import zipfile
import os
from glob import glob
from icecream import ic
import aiofiles
import asyncio
import time
from .loop import loop
import logging

async def download_images_data(organization: str, project: str, image_ids: List[str], chunk_size: int = 100) -> List[dict]:
    images_data = []
    starttime = time.time()
    for i in range(0, len(image_ids), chunk_size):
        chunk_ids = image_ids[i:i+chunk_size]
        async with loop.get(f'api/{organization}/projects/{project}/images?ids={",".join(chunk_ids)}') as response:
            assert response.status == 200, f'Error during downloading list of images. Statuscode is {response.status}'
            images_data += (await response.json())['images']
            logging.info(f'[+] Downloaded image data: {len(images_data)} / {len(image_ids)}')
            total_time = round(time.time() - starttime, 1)
            if(images_data):
                logging.debug(f'[+] Performance: {total_time} sec total. Per 100 : {(total_time / len(images_data ) * 100):.1f} sec')
            else:
                logging.debug(f'[+] Performance: {total_time} sec total.')
    return images_data


def create_resource_paths(organization_name: str, project_name: str, image_ids: List[str]) -> Tuple[List[str], List[str]]:
    if not image_ids:
        return [], []
    url_ids = [(f'api/{organization_name}/projects/{project_name}/images/{id}/main', id)
               for id in image_ids]
    urls, ids = list(map(list, zip(*url_ids)))

    return urls, ids


async def download_images(loop: asyncio.BaseEventLoop, paths: List[str], image_ids: List[str], image_folder: str, chunk_size: int = 10) -> None:
    starttime = time.time()
    for i in range(0, len(image_ids), chunk_size):
        chunk_paths = paths[i:i+chunk_size]
        chunk_ids = image_ids[i:i+chunk_size]
        tasks = []
        for j in range(len(chunk_paths)):
            tasks.append(loop.create_task(download_one_image(chunk_paths[j], chunk_ids[j], image_folder)))
        await asyncio.gather(*tasks)
        total_time = round(time.time() - starttime, 1)
        logging.info(f'Downnloaed {i + len(tasks)} image files.')
        logging.debug(f'[+] Performance (image files): {total_time} sec total. Per 100 : {(total_time / (i + len(tasks)) * 100):.1f}')


async def download_one_image(path: str, image_id: str, image_folder: str):
    async with loop.get(path) as response:
        assert response.status == 200, f'{response.status} for {path}'
        async with aiofiles.open(f'{image_folder}/{image_id}.jpg', 'wb') as out_file:
            await out_file.write(await response.read())


async def download_model(target_folder: str, organization: str, project: str, model_id: str, format: str) -> List[str]:
    # download model
    async with loop.get(f'api/{organization}/projects/{project}/models/{model_id}/{format}/file') as response:
        assert response.status == 200,  response.status
        provided_filename = response.headers.get("Content-Disposition").split("filename=")[1].strip('"')
        content = await response.read()

    # unzip and place downloaded model
    tmp_paht = f'/tmp/{os.path.splitext(provided_filename)[0]}'
    shutil.rmtree(tmp_paht, ignore_errors=True)
    filebytes = BytesIO(content)
    with zipfile.ZipFile(filebytes, 'r') as zip:
        zip.extractall(tmp_paht)

    created_files = []
    files = glob(f'{tmp_paht}/**/*', recursive=True)
    for file in files:
        logging.debug(f'moving model file {os.path.basename(file)} to training folder.')
        new_file = shutil.move(file, target_folder)
        created_files.append(new_file)
    return created_files


async def upload_model(organization: str, project: str, files: List[str], model_id: str, format: str) -> None:
    data = aiohttp.FormData()

    for file_name in files:
        data.add_field('files',  open(file_name, 'rb'))
    async with loop.put(f'api/{organization}/projects/{project}/models/{model_id}/{format}/file', data=data) as response:
        if response.status != 200:
            msg = f'---- could not save model with id {model_id}'
            raise Exception(msg)
        else:
            logging.info(f'---- uploaded model with id {model_id}')



from pydantic.main import BaseModel
from ..converter.converter import Converter
from ..status import State
from ..node import Node
from fastapi_utils.tasks import repeat_every
from icecream import ic
from ..loop import loop
import logging


class ModelInformation(BaseModel):
    organization: str
    project: str
    model_id: str


class ConverterNode(Node):
    converter: Converter
    skip_check_state: bool = False

    def __init__(self, name: str, uuid: str, converter: Converter):
        super().__init__(name, uuid)
        self.converter = converter

        @self.on_event("startup")
        @repeat_every(seconds=5, raise_exceptions=True, wait_first=False)
        async def check_state():
            if not self.skip_check_state:
                await self.check_state()

    async def convert_model(self, organization: str, project: str, model_id: str):
        await self.converter.convert(organization, project, model_id)
        await self.converter.upload_model(organization, project, model_id)

    async def check_state(self):
        logging.debug(f'checking state: {self.status.state}')

        if self.status.state == State.Running:
            return
        self.status.state = State.Running

        model = await self.find_model_to_convert()
        if model:
            await self.convert_model(model.organization, model.project, model.model_id)
        self.status.state = State.Idle

    async def find_model_to_convert(self) -> ModelInformation:
        async with loop.get('api/projects') as response:
            assert response.status == 200
            content = await response.json()
            projects = content['projects']

        for project in projects:
            organization_id = project['organization_id']
            project_id = project['project_id']
            path = f'api{project["resource"]}/models'
            async with loop.get(path) as models_response:
                assert models_response.status == 200
                content = await models_response.json()
                models = content['models']

                for model in models:
                    if model['version']:
                        if self.converter.source_format in model['formats'] and not self.converter.target_format in model['formats']:
                            return ModelInformation(organization=organization_id, project=project_id, model_id=model['id'])

    async def send_status(self):
        # NOTE not yet implemented
        pass



#
from pydantic.main import BaseModel
from typing import List
import asyncio
from icecream import ic
import os
from glob import glob
from ..context import Context
from .. import node_helper
from ..loop import loop
from ..node_helper import download_images_data, create_resource_paths
from ..trainer.basic_data import BasicData
from ..trainer.training_data import TrainingData
import logging


class DataDownloader(BaseModel):
    context: Context
    data_query_params: str

    async def download_data(self, image_folder: str) -> TrainingData:
        basic_data = await self.download_basic_data()
        training_data = await self.download_images_and_annotations(basic_data, image_folder)
        return training_data

    async def download_basic_data(self) -> BasicData:
        async with loop.get(f'api/{self.context.organization}/projects/{self.context.project}/data?{self.data_query_params}') as response:
            assert response.status == 200, response
            basic_data = BasicData.parse_obj(await response.json())
            return basic_data

    async def download_images_and_annotations(self, basic_data: BasicData, image_folder) -> TrainingData:
        loop = asyncio.get_event_loop()
        image_data_coroutine = self.download_image_data(basic_data.image_ids)

        image_data_task = loop.create_task(image_data_coroutine)
        await self.download_images(loop, basic_data.image_ids, image_folder)

        image_data = await image_data_task
        logging.info(f'Done downloading image_data for {len(image_data)} images.')
        return TrainingData(image_data=image_data, box_categories=basic_data.box_categories)

    async def download_image_data(self, ids: List[str]) -> List[dict]:
        return await download_images_data(self.context.organization, self.context.project, ids)

    async def download_images(self, loop, image_ids: List[str], image_folder: str) -> None:
        paths, ids = create_resource_paths(self.context.organization, self.context.project,
                                           self.filter_needed_image_ids(image_ids, image_folder))
        await node_helper.download_images(loop, paths, ids, image_folder)

    @staticmethod
    def filter_needed_image_ids(all_image_ids, image_folder) -> List[str]:
        ids = [os.path.splitext(os.path.basename(image))[0] for image in glob(f'{image_folder}/*.jpg')]
        return [id for id in all_image_ids if id not in ids]



from typing import List, Optional
import os
from pydantic.main import BaseModel
from learning_loop_node.context import Context
from learning_loop_node.node import Node
from icecream import ic
from learning_loop_node import node_helper


class Converter(BaseModel):
    model_folder: Optional[str]
    source_format: str
    target_format: str

    async def convert(self, organization: str, project: str, model_id: str) -> bool:
        project_folder = Node.create_project_folder(organization, project)

        self.model_folder = Converter.create_model_folder(project_folder, model_id)
        await node_helper.download_model(self.model_folder, organization,
                                   project, model_id, self.source_format)

        await self._convert()

    async def _convert(self) -> None:
        raise NotImplementedError()

    def get_converted_files(self, model_id) -> List[str]:
        raise NotImplementedError()

    async def upload_model(self, organization: str, project: str, model_id: str) -> bool:
        files = self.get_converted_files(model_id)
        await node_helper.upload_model(organization, project, files, model_id, self.target_format)

    @staticmethod
    def create_convert_folder(project_folder: str) -> str:
        image_folder = f'{project_folder}/images'
        os.makedirs(image_folder, exist_ok=True)
        return image_folder

    @staticmethod
    def create_model_folder(project_folder: str, model_id: str) -> str:
        model_folder = f'{project_folder}/{model_id}'
        os.makedirs(model_folder, exist_ok=True)
        return model_folder



from learning_loop_node.context import Context
from learning_loop_node.trainer.capability import Capability
from learning_loop_node.trainer.downloader import DataDownloader


class DownloaderFactory:

    @staticmethod
    def create(context: Context, capability: Capability) -> DataDownloader:
        if capability == Capability.Box:
            return DataDownloader(
                context=context,
                data_query_params='state=complete&mode=box'
            )
        else:
            raise NotImplementedError('Not implemented yet.')



from learning_loop_node.trainer.training_data import TrainingData
from pydantic import BaseModel
from typing import Optional
from learning_loop_node.trainer.model import Model
from learning_loop_node.context import Context


class Training(BaseModel):
    id: str
    context: Context

    project_folder: str
    images_folder: str

    training_folder: Optional[str]

    data: Optional[TrainingData]

    @property
    def last_known_model(self):
        if self.last_produced_model:
            return self.last_produced_model
        else:
            return self.base_model



from learning_loop_node.tests import test_helper
import pytest
import shutil


@pytest.fixture()
def create_project():
    test_helper.LiveServerSession().delete(f"/api/zauberzeug/projects/pytest?keep_images=true")
    project_configuration = {'project_name': 'pytest', 'inbox': 0, 'annotate': 0, 'review': 0, 'complete': 3, 'image_style': 'beautiful',
                             'categories': 2, 'thumbs': False, 'tags': 0, 'trainings': 1, 'detections': 3, 'annotations': 0, 'skeleton': False}
    assert test_helper.LiveServerSession().post(f"/api/zauberzeug/projects/generator",
                                                json=project_configuration).status_code == 200
    yield
    test_helper.LiveServerSession().delete(f"/api/zauberzeug/projects/pytest?keep_images=true")


@pytest.fixture(autouse=True, scope='function')
def cleanup_data_folder():
    shutil.rmtree('../data', ignore_errors=True)
    yield
    shutil.rmtree('../data', ignore_errors=True)



from learning_loop_node import DetectorNode
import pytest

@pytest.mark.asyncio
async def test_connect_without_credentials():
    node = DetectorNode(uuid='4ad7750b-4f0c-4d8d-86c6-c5ad04e19d2c', name='test detector')
    await node.connect() # should be able to connect without username/password



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['learning_loop_node',
 'learning_loop_node.converter',
 'learning_loop_node.detector',
 'learning_loop_node.tests',
 'learning_loop_node.trainer',
 'learning_loop_node.trainer.tests']

package_data = \
{'': ['*'], 'learning_loop_node.trainer.tests': ['test_data/*']}

install_requires = \
['aiofiles>=0.7.0,<0.8.0',
 'async_generator>=1.10,<2.0',
 'fastapi-socketio>=0.0.6,<0.0.7',
 'fastapi-utils>=0.2.1,<0.3.0',
 'fastapi>=0.63.0,<0.64.0',
 'icecream>=2.1.0,<3.0.0',
 'pytest-watch>=4.2.0,<5.0.0',
 'python-socketio[asyncio_client]>=5.0.4,<6.0.0',
 'requests>=2.25.1,<3.0.0',
 'simplejson>=3.17.2,<4.0.0',
 'uvicorn>=0.13.3,<0.14.0',
 'werkzeug>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'learning-loop-node',
    'version': '0.2.1',
    'description': 'Python Library for Nodes which connect to the Zauberzeug Learning Loop',
    'long_description': '# Learning Loop Node\n\nThis Python library helps you to write your own Detection Nodes, Training Nodes and Converter Nodes for the Zauberzeug Learning Loop.\n\n## General Usage\n\nYou can configure connection to the learning loop by specifying the following environment variables before starting:\n\n- HOST=learning-loop.ai\n- ORGANIZATION=<your organization>\n- PROJECT=<your project>\n- USERNAME=<your username>\n- PASSWORD=<your password>\n\n## Detector Node\n\n## Trainer Node\n\n## Converter Node\n\nA Conveter Node converts models from one format into another.\n\n### How to test the operability?\n\nAssumend there is a Converter Node which converts models of format \'format_a\' into \'format_b\'.\nUpload a model with\n`curl --request POST -F \'files=@my_model.zip\' https://learning-loop.ai/api/zauberzeug/projects/demo/format_a`\nThe model should now be available for the format \'format_a\'\n`curl "https://learning-loop.ai/api/zauberzeug/projects/demo/models?format=format_a"`\n\n```\n{\n  "models": [\n    {\n      "id": "3c20d807-f71c-40dc-a996-8a8968aa5431",\n      "version": "4.0",\n      "formats": [\n        "format_a"\n      ],\n      "created": "2021-06-01T06:28:21.289092",\n      "comment": "uploaded at 2021-06-01 06:28:21.288442",\n      ...\n    }\n  ]\n}\n\n```\n\nbut not in the format_b\n`curl "https://learning-loop.ai/api/zauberzeug/projects/demo/models?format=format_b"`\n\n```\n{\n  "models": []\n}\n```\n\nConnect the Node to the learning loop by simply starting the container.\nAfter a short time the converted Model should be available as well.\n`curl https://learning-loop.ai/api/zauberzeug/projects/demo/models?format=format_b`\n\n```\n{\n  "models": [\n    {\n      "id": "3c20d807-f71c-40dc-a996-8a8968aa5431",\n      "version": "4.0",\n      "formats": [\n        "format_a",\n        "format_b",\n      ],\n      "created": "2021-06-01T06:28:21.289092",\n      "comment": "uploaded at 2021-06-01 06:28:21.288442",\n      ...\n    }\n  ]\n}\n```\n',
    'author': 'Zauberzeug GmbH',
    'author_email': 'info@zauberzeug.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zauberzeug/learning_loop_node',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
